"""
Created on 27 Sep 2020

:author: semuadmin (Steve Smith)
:copyright: 2020 semuadmin
:license: BSD 3-Clause
"""

# pylint: disable=invalid-name

from pygpsclient._version import __version__
from pygpsclient.helpers import nmea2preset, ubx2preset

version = __version__
